/*
 * file.c
 *
 * All rights reserved. Copyright (C) 1994,1998 by NARITA Tomio
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef MSDOS
#include <dos.h>
#endif /* MSDOS */

#ifdef UNIX
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#endif /* UNIX */

#include <import.h>
#include <decode.h>
#include <encode.h>
#include <console.h>
#include <uty.h>
#include <guess.h>
#include <begin.h>
#include <file.h>

public boolean_t FileLoadLine( file_t *f, int *length )
{
  boolean_t flagSimple = TRUE;
  boolean_t flagEightBit = FALSE, flagHz = FALSE;
  int i, ch;

  i = 0;
  while( EOF != (ch = fgetc( f->fp )) ){
    load_str[ i++ ] = (char)ch;
    if( LF == ch )
      break;
    if( i == ( STR_SIZE - 1 ) )
      break;
    if( FALSE == flagEightBit ){
      if( ch > DEL ){
	flagEightBit = TRUE;
	flagSimple = FALSE;
      } else if( FALSE == flagHz ){
	if( '~' == ch && TRUE == hz_detection ){
	  flagHz = TRUE;
	  flagSimple = FALSE;
	}
      }
    }
    if( TRUE == flagSimple ){
      if( ( ch < SP && CR != ch ) || DEL == ch
	 || ( '+' == ch && UTF_7 == f->inputCodingSystem ) )
	flagSimple = FALSE;
    }
  }

  *length = i;

  if( EOF == ch )
    f->eof = TRUE;
  else
    f->eof = FALSE;

  if( AUTOSELECT == f->inputCodingSystem ){
    if( TRUE == flagEightBit ){
      f->inputCodingSystem = GuessCodingSystem( load_str, i,
					       f->defaultCodingSystem );
      if( NULL != f->find.pattern )
	AdjustPatternCharset( f->inputCodingSystem, f->keyboardCodingSystem,
			     f->defaultCodingSystem, f->find.pattern );
    } else if( TRUE == flagHz ){
      if( GuessHz( load_str, i ) ){
	f->inputCodingSystem = HZ_GB;
	if( NULL != f->find.pattern )
	  AdjustPatternCharset( f->inputCodingSystem, f->keyboardCodingSystem,
			       f->defaultCodingSystem, f->find.pattern );
      }
    }
  }

  return flagSimple;
}

public boolean_t FileStretch( file_t *f, unsigned int target )
{
  int ch, count;
  unsigned int segment, line;
  long ptr;

  if( TRUE == f->done )
    return FALSE;

  line = 0;
  count = 0;
  ptr = f->lastPtr;
  segment = f->lastSegment;

  if( fseek( f->fp, ptr, SEEK_SET ) )
    perror( "FileStretch()" ), exit( -1 );

#ifndef MSDOS /* IF NOT DEFINED */
  if( NULL != f->sp ){
    while( EOF != (ch = fgetc( f->sp )) ){
      fputc( ch, f->fp );
      if( TRUE == immediate_print ){
	putchar( ch );
	f->dirty = TRUE;
      }
      count++;
      if( LF == ch || count == ( STR_SIZE - 1 ) ){
	count = 0;
	if( 0 > (ptr = ftell( f->fp )) )
	  perror( "FileStretch()" ), exit( -1 );
	if( ++line == PAGE_SIZE ){
	  f->totalLines += line;
	  line = 0;
	  if( 0 == Slot( ++segment ) ){
	    if( FRAME_SIZE == ++f->lastFrame
	       ||
	       NULL == (f->slot[ f->lastFrame ]
			= (long *)malloc( sizeof( long ) * SLOT_SIZE ))
	       ){
	      f->done = TRUE;
	      f->truncated = TRUE;
	      return FALSE;
	    }
	  }
	  f->slot[ f->lastFrame ][ Slot( segment ) ] = ptr;
	  f->lastSegment = segment;
	  f->lastPtr = ptr;
	  if( segment > target )
	    return TRUE;
	}
	if( TRUE == kb_interrupted )
	  return FALSE;
      }
    }
  } else {
#endif /* MSDOS */
    while( EOF != (ch = fgetc( f->fp )) ){
      count++;
      if( LF == ch || count == ( STR_SIZE - 1 ) ){
	count = 0;
	if( 0 > (ptr = ftell( f->fp )) )
	  perror( "FileStretch()" ), exit( -1 );
	if( ++line == PAGE_SIZE ){
	  f->totalLines += line;
	  line = 0;
	  if( 0 == Slot( ++segment ) ){
	    if( FRAME_SIZE == ++f->lastFrame
	       ||
	       NULL == (f->slot[ f->lastFrame ]
			= (long *)malloc( sizeof( long ) * SLOT_SIZE ))
	       ){
	      f->done = TRUE;
	      f->truncated = TRUE;
	      return FALSE;
	    }
	  }
	  f->slot[ f->lastFrame ][ Slot( segment ) ] = ptr;
	  f->lastSegment = segment;
	  f->lastPtr = ptr;
	  if( segment > target )
	    return TRUE;
	}
#ifdef MSDOS
	if( TRUE == allow_interrupt )
	  bdos( 0x0b, 0, 0 );
#endif /* MSDOS */
	if( TRUE == kb_interrupted )
	  return FALSE;
      }
    }
#ifndef MSDOS /* IF NOT DEFINED */
  }
#endif /* MSDOS */

  if( FALSE == kb_interrupted ){
    if( 0 < line || 0 < count ){
      segment++;
      f->totalLines += line;
      f->lastSegment = segment;
      if( 0 > (f->lastPtr = ftell( f->fp )) )
	perror( "FileStretch()" ), exit( -1 );
    }
    f->done = TRUE;
  }

  if( segment > target )
    return TRUE;
  else
    return FALSE;
}

public boolean_t FileSeek( file_t *f, unsigned int segment )
{
  if( segment >= f->lastSegment )
    if( FALSE == FileStretch( f, segment ) )
      return FALSE;

  if( fseek( f->fp, f->slot[ Frame( segment ) ][ Slot( segment ) ], SEEK_SET ) )
    perror( "FileSeek()" ), exit( -1 );

  return TRUE;
}

public void FileFreeHead( file_t *f )
{
  int i, j;

  for( i = 0 ; i < BLOCK_SIZE ; i++ ){
    for( j = 0 ; j < PAGE_SIZE ; j++ ){
      if( NULL != f->page[ i ].line[ j ].head )
	free( f->page[ i ].line[ j ].head );
      f->page[ i ].line[ j ].head = NULL;
    }
  }
}

public void FileRefresh( file_t *f )
{
  int i, j;

  for( i = 0 ; i < BLOCK_SIZE ; i++ ){
    f->used[ i ] = FALSE;
    f->page[ i ].lines = 0;
    for( j = 0 ; j < PAGE_SIZE ; j++ ){
      f->page[ i ].line[ j ].heads = 0;
      f->page[ i ].line[ j ].istr = NULL;
      f->page[ i ].line[ j ].head = NULL;
    }
  }
}

public file_t *FileOpen( char *fileName, FILE *fp, FILE *sp,
			int width, int height,
			char inputCodingSystem,
			char outputCodingSystem,
			char keyboardCodingSystem,
			char pathnameCodingSystem,
			char defaultCodingSystem )
{
  file_t *f;
  int i;

#ifdef MSDOS
#undef file_t
  if( NULL == (f = (file_t far *)FarMalloc( sizeof( file_t ) )) )
    NotEnoughMemory();
#define file_t file_t far
#else
  f = (file_t *)Malloc( sizeof( file_t ) );
#endif /* MSDOS */

  f->fileName		= fileName;
  f->fileNameI18N	= NULL;
  f->fileNameLength	= strlen( fileName );
  f->fp			= fp;
  f->sp			= sp;
  f->lastSegment	= 0;
  f->totalLines		= 0L;
  f->lastPtr		= 0L;

  for( i = 0 ; i < FRAME_SIZE ; i++ )
    f->slot[ i ] = NULL;

  f->slot[ 0 ]		= (long *)Malloc( sizeof( long ) * SLOT_SIZE );
  f->slot[ 0 ][ 0 ]	= 0L;
  f->lastFrame		= 0;

  f->done		= FALSE;
  f->eof		= FALSE;
  f->top		= TRUE;
  f->bottom		= FALSE;
  f->truncated		= FALSE;
  f->dirty		= FALSE;

  f->find.pattern	= NULL;
  f->find.displayed	= FALSE;

  f->width		= width;
  f->height		= height;

  f->inputCodingSystem		= inputCodingSystem;
  f->outputCodingSystem		= outputCodingSystem;
  f->keyboardCodingSystem	= keyboardCodingSystem;
  f->pathnameCodingSystem	= pathnameCodingSystem;
  f->defaultCodingSystem	= defaultCodingSystem;

  FileRefresh( f );
  FileStretch( f, 0 );

  return f;
}

public boolean_t FileClose( file_t *f )
{
  int i;

  if( NULL == f )
    return FALSE;

  for( i = 0 ; i < FRAME_SIZE ; i++ ){
    if( NULL != f->slot[ i ] )
      free( f->slot[ i ] );
  }

  FileFreeHead( f );

#ifdef MSDOS
  FarFree( f );
#else
  free( f );
#endif /* MSDOS */

  return TRUE;
}

public char *FileName( file_t *f )
{
  int length;

  if( NULL == f->fileNameI18N ){
    length = f->fileNameLength;
    f->fileNameI18N = Decode( f->pathnameCodingSystem,
			     f->fileName,
			     &length );
    f->fileNameLength = length;		/* for MSDOS */
  }

  Encode( f->fileNameI18N, 0, f->fileNameLength,
	 f->outputCodingSystem, FALSE,
	 encode_str, &encode_length );

  return EncodeStripAttribute( encode_str, encode_length );
}

private char fileStatus[ 256 ];

public char *FileStatus( file_t *f )
{
  sprintf( fileStatus, "%s %lu/%lu [%s|%s|%s|%s]",
	  FileName( f ),
	  (unsigned long)( 1 + f->screen.top.seg * PAGE_SIZE + f->screen.top.off ),
	  f->totalLines,
	  cTable[ (int)f->inputCodingSystem ].codingSystemName,
	  cTable[ (int)f->keyboardCodingSystem ].codingSystemName,
	  cTable[ (int)f->outputCodingSystem ].codingSystemName,
	  cTable[ (int)f->pathnameCodingSystem ].codingSystemName
	  );

  return fileStatus;
}
